----------------------------------------------------------------------------------
-- Project Name: Gage Hash Function, C: 224 and r: 8
-- 
-- Copyright 2019:
--     Mohamed El-Hadedy <mealy@cpp.edu>
--     Danilo Gligoroski <danilog@ntnu.no>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-- 02110-1301, USA.
-- 
----------------------------------------------------------------------------------
----------------------------------------------------------------------------------
-- Company: California State Polytechnic University, Pomona 
-- Engineer: Assistant Professor, ECE-department, College of Engineering 
-- 
-- Create Date: 07/10/2019 04:57:47 PM
-- Design Name: Mohamed El-Hadedy 
-- Module Name: SMART_ENGINE - Behavioral
-- Project Name: Gage Hash function 
-- Target Devices: Basys3
-- Tool Versions: Vivado 2019.1
-- Description: Smart engine is responsible on computing the main part of the
-- hashing operations 
-- 
-- Dependencies: gage1h256c224r008AllSequential, 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.all;
use ieee.STD_LOGIC_ARITH.all;
use ieee.std_logic_unsigned.all;
use IEEE.math_real.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity gage1h256c224r008AllSequential is
   generic (C: positive:= 224; r: positive:=8); 
    Port ( clk     : in  STD_LOGIC;
           rst     : in  STD_LOGIC;
           M_ST    : in  STD_LOGIC; 
           wr_en   : in  STD_LOGIC;
           Message : in  STD_LOGIC_VECTOR (7 downto 0);
           HASH_MSG: out STD_LOGIC_VECTOR (7 downto 0);
           HASH_OUT: out STD_LOGIC
         );
end gage1h256c224r008AllSequential;

architecture Behavioral of gage1h256c224r008AllSequential is

component Engine_Q 
    Port ( clk       : in  STD_LOGIC;
           rst       : in  STD_LOGIC;
           STRT      : in  STD_LOGIC; 
           L         : in  STD_LOGIC_VECTOR(1 downto 0);
           OSTATE    : in  STD_LOGIC;
           INPQ      : out STD_LOGIC;
           QEN_OUT   : out STD_LOGIC;
           RAM_WR    : out STD_LOGIC 
         );
end component;



component OST_RAM
    generic (c:positive:=224 ; r:positive:= 8);
    Port (
            clk   : in  std_logic;
            RAM_wr: in  std_logic;  
            PTR_wr: in  std_logic_vector(integer(ceil(log2(real(c+r))))-1 downto 0); 
            ST_INP: in  std_logic; 
            RAM_rd: in  std_logic; 
            PTR_rd: in  std_logic_vector(integer(ceil(log2(real(c+r))))-1 downto 0); 
            ST_OP : out std_logic 
          );
end component;

component Leader 
    Port ( LDR_INP : in STD_LOGIC_VECTOR (4 downto 0);
           LDR_OP : out STD_LOGIC_VECTOR (1 downto 0));
end component;

component STD_FIFO 
	Generic (
		constant DATA_WIDTH  : positive := 8;
		constant FIFO_DEPTH	: positive := 8
	);
	Port ( 
		CLK		: in  STD_LOGIC;
		RST		: in  STD_LOGIC;
		WriteEn	: in  STD_LOGIC;
		DataIn	: in  STD_LOGIC_VECTOR (DATA_WIDTH - 1 downto 0);
		ReadEn	: in  STD_LOGIC;
		DataOut	: out STD_LOGIC_VECTOR (DATA_WIDTH - 1 downto 0);
		Empty	: out STD_LOGIC;
		Full	: out STD_LOGIC
	);
end component;   


component BIT_SHUFFLER
    generic (c:positive:=224 ; r:positive:= 8);

    Port ( clk    : in  STD_LOGIC;
           rst    : in  STD_LOGIC; 
           OS_POS : in  STD_LOGIC_VECTOR (integer(ceil(log2(real(c+r))))-1 downto 0);
           NS_POS : out STD_LOGIC_VECTOR (integer(ceil(log2(real(c+r))))-1 downto 0)
         );
end component;


-- SMART SIGNALS --

signal SMART_main_SEL: std_logic; 
signal SMART_sec_SEL : std_logic; 

-- Main BUFFER --
signal RAM_wr_main: std_logic;
signal PTR_wr_main: std_logic_vector(integer(ceil(log2(real(c+r))))-1 downto 0);
signal ST_INP_main: std_logic; 
signal RAM_rd_main: std_logic;
signal PTR_rd_main: std_logic_vector(integer(ceil(log2(real(c+r))))-1 downto 0);
signal ST_OP_main : std_logic; 
signal BUFF_XOR   : std_logic; 

-- SECONDARY BUFFER --
signal RAM_wr_sec: std_logic;
signal PTR_wr_sec: std_logic_vector(integer(ceil(log2(real(c+r))))-1 downto 0);
signal ST_INP_sec: std_logic; 
signal RAM_rd_sec: std_logic;
signal PTR_rd_sec: std_logic_vector(integer(ceil(log2(real(c+r))))-1 downto 0);
signal ST_OP_sec : std_logic; 

-- BUFF ASSOC WITH Q ENGINE  --
signal BUFF_SEL   : std_logic;
signal MSG_SEL    : std_logic;  
Signal ST_OP_Q_INP: std_logic; 

signal INPQ_SEL   : std_logic_vector(1 downto 0); 
-- Q_SBOX signals -- 
signal INPQ: std_logic_vector(3 downto 0); 
signal Q_OP: std_logic; 
signal QEN_OUT: std_logic;
signal SMRT_INP:std_logic; 

-- Leader signals --
signal LDR_INP : std_logic_vector(4 downto 0); 
signal L       : std_logic_vector(1 downto 0); 
signal INIT_ST : std_logic; 
signal INP_ST  : std_logic; 
signal MSG_XOR : std_logic; 
signal MSG_POS : std_logic_vector(2 downto 0); 
signal Q_STRT  : std_logic; 


signal MSG       : STD_LOGIC_VECTOR(7 downto 0);
signal START     : STD_LOGIC; 
signal SMRT_ENOUT: STD_LOGIC;
signal empty     : std_logic; 
signal full      : std_logic; 
signal rd_en     : std_logic; 
signal MSG_FINL  : std_logic_vector(7 downto 0); 
constant FINL    : std_logic_vector(7 downto 0):= "10000000";
signal FIN_SEL   : std_logic;  


signal main_sel   : std_logic; 
signal sec_sel    : std_logic; 
signal OS_POS     : std_logic_vector(7 downto 0);
signal NS_POS     : std_logic_vector(7 downto 0);
signal PTR_wr_init: std_logic_vector(7 downto 0); 

signal intrn_count: std_logic_vector(7 downto 0); 
signal V          : std_logic_vector(3 downto 0); 
signal T          : std_logic_vector(1 downto 0);
signal rst_auto, rst_engine:std_logic; 
signal RAM_Q      : std_logic; 
signal TAG        : std_logic_vector(5 downto 0); 
signal HASH_POS   : std_logic_vector(2 downto 0);

--signal COUNT_OUT  : std_logic_vector(4 downto 0); 
-- State Machine --
  type fsm_states is (
                      INIT,
                      CLR_MAIN,
              --        FIN_PROG,
                      INJECTMSG,
                      LDMAIN,
                      LDSEC,
                      FINST_OUTPUT,
                      OUTPUT
                     );
  -- signals declaration
  signal reg_fsm, nxt_fsm   : fsm_states;
  
  type fsm_output is (
                       INIT_OUT,
                       SEC_OUT, 
                       MAIN_OUT,
                       OUTOUT,
                       DONE 
                     ); 
  signal reg_fsm_out, nxt_fsm_out: fsm_output;                    
begin


FIFO_ENG: STD_FIFO generic map (DATA_WIDTH => 8, FIFO_DEPTH => 32)
                   	Port   map ( 
                                CLK		=> clk,
                                RST		=> rst,
                                WriteEn	=> wr_en,
                                DataIn	=> Message,
                                ReadEn	=> rd_en,
                                DataOut	=> MSG,
                                Empty	=> empty,
                                Full	=> full 
                            
                              );       


INP_ST      <= ((MSG_SEL and BUFF_XOR)    or (not MSG_SEL and INIT_ST));
ST_INP_main <= ((SMART_main_SEL and Q_OP) or (not SMART_main_SEL and INP_ST)); 

MAIN_BUFFER: OST_RAM generic map (c => c, r =>r)
                 port map (
                           clk    => clk, 
                           RAM_wr => RAM_wr_main,
                           PTR_wr => PTR_wr_main,
                           ST_INP => ST_INP_main,
                           RAM_rd => RAM_rd_main,
                           PTR_rd => PTR_rd_main, 
                           ST_OP  => ST_OP_main
                          ); 

SECONDARY_BUFFER: OST_RAM generic map (c => c, r =>r)
                 port map (
                           clk    => clk, 
                           RAM_wr => RAM_wr_sec,
                           PTR_wr => PTR_wr_sec,
                           ST_INP => ST_INP_sec,
                           RAM_rd => RAM_rd_sec,
                           PTR_rd => PTR_rd_sec,
                           ST_OP  => ST_OP_sec
                          ); 
                          
                          
 
ST_INP_sec <= ((SMART_sec_SEL and Q_OP) or (not SMART_sec_SEL and INIT_ST)); 

                          
 GEN_FOR: for i in 0 to r-1 generate 
    MSG_FINL(r-1-i) <= (FIN_SEL and FINL(r-1-i)) or (not FIN_SEL and MSG(r-1-i)); 
 end generate; 
 
 
 GEN_LVL_1_MUX: for i in 0 to 3 generate
    V(i) <= (not MSG_POS(2) and MSG_FINL(7-i)) or (MSG_POS(2) and MSG_FINL(3-i));
 end generate;
                          
 GEN_LVL_2_MUX: for i in 0 to 1 generate
    T(i) <= (not MSG_POS(1) and V(i)) or (MSG_POS(1) and V(2+i));
 end generate;
 
    MSG_XOR <= (not MSG_POS(0) and T(0)) or (MSG_POS(0) and T(1));
                          
                          
  BUFF_XOR  <= ST_OP_main xor MSG_XOR;           
            
--ST_OP_main <= ((MSG_SEL and (ST_OP_main_BUFF xor MSG_XOR)) or (not MSG_SEL and ST_OP_main_BUFF)); 
                          

ST_OP_Q_INP <= ((BUFF_SEL and ST_OP_sec) or (not BUFF_SEL and ST_OP_main)); 
             
LDR_PROCESS: Leader port map ( 
                                LDR_INP => LDR_INP,
                                LDR_OP  => L    
                             );                         
rst_auto <= rst or rst_engine; 
ENGININE_BF_Q: Engine_Q port map 
                                ( clk      => clk ,
                                  rst      => rst_auto , 
                                  STRT     => Q_STRT,
                                  L        => L   , 
                                  OSTATE   => ST_OP_Q_INP,
                                  INPQ     => Q_OP,
                                  QEN_OUT  => QEN_OUT,
                                  RAM_WR   => RAM_Q
                                );
                                
GEN_ADR_MAIN: for i in 0 to r-1 generate 
    PTR_wr_main(r-1-i) <= (main_sel and NS_POS(r-1-i)) or (not main_sel and PTR_wr_init(r-1-i)); 
 end generate; 
GEN_ADR_SEC: for i in 0 to r-1 generate 
    PTR_wr_sec(r-1-i) <= (sec_sel and NS_POS(r-1-i)) or (not sec_sel and PTR_wr_init(r-1-i)); 
 end generate;   
 
SHUFF: BIT_SHUFFLER  generic map(c=>c , r=> r)
                     Port    map(
                                 clk    => clk,
                                 rst    => rst, 
                                 OS_POS => OS_POS,
                                 NS_POS => NS_POS
                                );
 
                     
--Finite State Machine --

-- state machine
  dff_fsm: process(rst, clk)
  begin
    if (rst = '1') then
      reg_fsm     <= INIT;
    elsif (rising_edge(clk)) then
      reg_fsm <= nxt_fsm;
    end if;
  end process dff_fsm;
  
 cmb_fsm: process(clk, rst,reg_fsm, M_ST)
  begin
   if rst = '1' then 
       HASH_POS    <=  (others =>'0');
       HASH_OUT    <=  '0';
       MSG_POS     <=  (others =>'0'); 
       intrn_count <=  (others =>'0'); 
       main_sel    <=  '1';
       sec_sel     <=  '1';
       Q_STRT      <=  '0';
       MSG_SEL     <=  '0'; 
       SMART_main_SEL <=  '0';
       SMART_sec_SEL  <=  '0';
       rd_en       <= '0';
       INIT_ST     <=  '0';
       RAM_WR_main <=  '0';
       RAM_RD_main <=  '0'; 
       PTR_RD_main <= (others =>'0'); 
       RAM_WR_sec  <=  '0';
       RAM_RD_sec  <=  '0'; 
       PTR_RD_sec  <= (others =>'0'); 
       BUFF_SEL    <= '0';
       LDR_INP     <= (others =>'0');   
       INPQ_SEL    <= (others =>'0');
       SMRT_ENOUT  <= '0'; 
       FIN_SEL     <= '0';
       rst_engine  <= '0';
       TAG         <= (others =>'0');
      -- COUNT_OUT   <= (others =>'0'); 
   elsif rising_edge (clk) then  
    case reg_fsm is
      when INIT => 
         if (M_ST = '1') then
           nxt_fsm <= CLR_MAIN;
           main_sel    <=  '0';
           sec_sel     <=  '0'; 
           RAM_WR_main <=  '1';
           RAM_WR_sec  <=  '1';    
           MSG_SEL     <=  '0';
           rst_engine  <=  '1';
         else 
            nxt_fsm  <= INIT; 
          --  COUNT_OUT<=(others=>'0');
          tag <=(others=>'0');
         end if;
      when CLR_MAIN =>
                if (PTR_wr_init >= 231) then 
                    RAM_wr_main <= '0';
                    RAM_wr_sec  <= '0';
                     if empty = '1' then 
                        FIN_SEL  <= '1';
                        tag      <= "000001";
                        nxt_fsm  <= INJECTMSG; 
                     else 
                        tag      <= "000000";
                        FIN_SEL  <= '0';
                        nxt_fsm <= INJECTMSG; 
                     end if;
                --     nxt_fsm<= INJECTMSG;  
                end if;  
             --   rst_engine   <= '1';
--      when FIN_PROG  =>
--        if tag = 0 then 
--           nxt_fsm <= INJECTMSG;
--           FIN_SEL <= '1';
--        elsif tag = 2 then  
--           FIN_SEL <= '1';
--           intrn_count <= x"01"; 
--           PTR_rd_main <= (others =>'0');
--           RAM_rd_main <= '1';
--           nxt_fsm <= INJECTMSG;
--        end if; 
      when INJECTMSG =>
           rst_engine   <= '1';
          if  intrn_count = 0 then   
            if empty = '1' then 
                FIN_SEL <= '1';
            else 
                FIN_SEL <= '0';
            end if; 
            main_sel    <= '0';
            intrn_count <= intrn_count + 1;
          elsif intrn_count = 1  then
            main_sel <='0';
               if empty = '1' then  
                 rd_en       <= '0';
              --   FIN_SEL     <= '1';
                 RAM_rd_main <= '0';
                else 
            --     FIN_SEL     <= '0';
                 rd_en       <= '1';
                 RAM_rd_main <= '1';
                end if; 
             MSG_SEL     <= '1';
             SMART_main_SEL <= '0';
             MSG_POS     <= (others =>'0'); 
             intrn_count <= intrn_count + 1;
           elsif(intrn_count = 2) then
            rd_en        <= '0'; 
            RAM_rd_main  <= '1';
            intrn_count <= intrn_count + 1;
           elsif(intrn_count = 3) then 
             MSG_POS     <= "000";
             PTR_rd_main <= (others =>'0');
             RAM_wr_main <= '1';
             intrn_count <= intrn_count+1; 
           elsif(intrn_count = 4) then 
            RAM_wr_main  <= '0';
            intrn_count  <= intrn_count+1;
           elsif(intrn_count = 5) then 
             MSG_POS      <= MSG_POS + 1;
             PTR_rd_main  <= PTR_rd_main+1;
             RAM_wr_main <= '0';
             intrn_count <= intrn_count+1; 
           elsif(intrn_count = 6) then 
               if (MSG_POS =0) then 
                  RAM_wr_main  <= '0';
                  intrn_count  <= intrn_count+1;   
               else 
                RAM_wr_main <=  '1';
                intrn_count  <= intrn_count -1;  
               end if;   
           elsif(intrn_count = 7) then 
               RAM_wr_main <= '0';
               MSG_POS     <= (others =>'0');
               PTR_rd_main <= (others =>'0');
               RAM_rd_main <= '0';
               intrn_count <= (others =>'0');
               LDR_INP     <= (others =>'0');
               BUFF_SEL    <= '0'; 
               SMART_sec_SEL <= '1';
               Q_STRT        <= '0';
               nxt_fsm     <= LDMAIN; 
           end if;   
           
      when LDMAIN =>
             SMART_sec_SEL  <= '1';
             SMART_main_SEL <= '0';
             BUFF_SEL    <= '0'; 
             sec_sel     <= '1'; 
             main_sel    <= '0';
           if intrn_count = 1  then
             rst_engine  <= '0';
             Q_STRT      <= '1'; 
             intrn_count <= intrn_count + 1;
           elsif(intrn_count = 2) then
             Q_STRT      <= '0';
             RAM_rd_main <= '1';
             PTR_rd_main <= (x"00"); 
             intrn_count <= intrn_count + 1;
             RAM_rd_main <= '1';
          elsif (intrn_count = 3) then 
             RAM_rd_main <= '1';
             PTR_rd_main <= PTR_rd_main+1;
             intrn_count <= intrn_count + 1; 
          elsif (intrn_count = 4) then 
             RAM_rd_main <= '0';
          if (PTR_rd_main <231) then 
             intrn_count <= intrn_count + 1;
          else  
             intrn_count <= intrn_count + 3;
          end if; 
          elsif (intrn_count = 5) then 
             RAM_rd_main <= '1';
             PTR_rd_main <= PTR_rd_main+1;
             intrn_count <= intrn_count + 1;
         elsif (intrn_count = 6) then 
             RAM_rd_main <= '1';
             PTR_rd_main <= PTR_rd_main+1;
             intrn_count <= intrn_count - 2;     
         elsif(intrn_count = 7) then 
            intrn_count <= intrn_count +1 ;
         elsif(intrn_count = 8) then 
            intrn_count <= intrn_count +1;
         elsif(intrn_count = 9) then 
            intrn_count <= intrn_count + 1;
         elsif(intrn_count = 10) then 
            RAM_rd_main  <= '0';
            intrn_count  <= (others =>'0');
            PTR_rd_main  <= (others =>'0');
            rst_engine   <= '1';
            LDR_INP      <= LDR_INP + 1;
            nxt_fsm      <= LDSEC;
          end if;
             RAM_wr_SEC <= RAM_Q;          
      when LDSEC  =>
             SMART_main_SEL<= '1';
             SMART_sec_SEL <= '0';
             BUFF_SEL      <= '1'; 
             sec_sel       <= '0'; 
            if LDR_INP <31 then 
             main_sel      <= '1';
            else 
             main_sel      <= '0';
            end if; 
           if intrn_count = 0  then 
            intrn_count <= intrn_count + 1;
           elsif intrn_count = 1  then
             rst_engine  <= '0';
             Q_STRT      <= '1'; 
             intrn_count <= intrn_count + 1;
           elsif(intrn_count = 2) then
             Q_STRT      <= '0';
             RAM_rd_sec  <= '1';
             PTR_rd_sec  <= (x"00"); 
             intrn_count <= intrn_count + 1;
          elsif (intrn_count = 3) then 
             RAM_rd_sec  <= '1';
             PTR_rd_sec  <= PTR_rd_sec+1;
             intrn_count <= intrn_count + 1; 
          elsif (intrn_count = 4) then 
             RAM_rd_sec  <= '0';
          if (PTR_rd_sec <231) then 
             intrn_count <= intrn_count + 1;
          else  
             intrn_count <= intrn_count + 3;
          end if; 
          elsif (intrn_count = 5) then 
             RAM_rd_sec <= '1';
             PTR_rd_sec <= PTR_rd_sec+1;
             intrn_count <= intrn_count + 1;
         elsif (intrn_count = 6) then 
             RAM_rd_sec <= '1';
             PTR_rd_sec <= PTR_rd_sec+1;
             intrn_count <= intrn_count - 2;     
         elsif(intrn_count = 7) then 
            intrn_count <= intrn_count +1 ;
         elsif(intrn_count = 8) then 
            intrn_count <= intrn_count +1;
         elsif(intrn_count = 9) then 
            intrn_count <= intrn_count + 1; 
         elsif(intrn_count = 10) then 
            RAM_rd_sec   <= '0';
            PTR_rd_sec   <= (others =>'0');
            intrn_count  <= (others =>'0');
            rst_engine   <= '1';
            LDR_INP      <= LDR_INP + 1;
            if PTR_wr_init >= 231 then 
               intrn_count  <= (others =>'0');

                  if empty = '1' then 
                 --  if tag = 0 then 
                 --   RAM_rd_main <= '1';
               --     elsif tag = 1 then  
               --     nxt_fsm <= FIN_PROG;
               --     intrn_count  <= (others =>'0');
                --    end if;  
                    if tag >=1 then 
                        nxt_fsm <= FINST_OUTPUT;
                        FIN_SEL      <='0';
                        intrn_count  <= (others =>'0');
                        RAM_rd_main <= '1';
                    else 
                       nxt_fsm <= INJECTMSG; 
                       intrn_count  <= (others =>'0');
                       FIN_SEL <= '1';
                    end if; 
                    tag     <= tag+1;

                  else 
                    nxt_fsm <= INJECTMSG; 
                    FIN_SEL <= '0';
                  end if; 
              else 
                nxt_fsm      <= LDMAIN;
             end if;                   
          end if;
             RAM_wr_MAIN <= RAM_Q;    
             
            
      when FINST_OUTPUT  =>
            LDR_INP     <= (others=>'0');
            if intrn_count = 0 then 
            PTR_rd_main <= (others =>'0');
            HASH_POS    <= (others =>'0');
         --   tag         <= (others =>'0');
            intrn_count <= intrn_count + 1;
            elsif (intrn_count = 1) then
            RAM_rd_main <= '1';
            HASH_POS    <= HASH_POS;
            PTR_rd_main <= PTR_rd_main+1;
            intrn_count <= intrn_count + 1;
            elsif (intrn_count = 2) then
            RAM_rd_main <= '1';
            HASH_POS    <= HASH_POS+1;
            PTR_rd_main <= PTR_rd_main+1;
            intrn_count <= intrn_count + 1;  
            elsif (intrn_count = 3) then
            RAM_rd_main <= '1';
            HASH_POS    <= HASH_POS+1;
            PTR_rd_main <= PTR_rd_main+1;
            intrn_count <= intrn_count + 1;  
           elsif (intrn_count = 4) then
            RAM_rd_main <= '1';
            HASH_POS    <= HASH_POS+1;
            PTR_rd_main <= PTR_rd_main+1;
            intrn_count <= intrn_count + 1;              
           elsif (intrn_count = 5) then
            RAM_rd_main <= '1';
            HASH_POS    <= HASH_POS+1;
            PTR_rd_main <= PTR_rd_main+1;
            intrn_count <= intrn_count + 1; 
           elsif (intrn_count = 6) then
            RAM_rd_main <= '1';
            HASH_POS    <= HASH_POS+1;
            PTR_rd_main <= PTR_rd_main+1;
            intrn_count <= intrn_count + 1;               
           elsif (intrn_count = 7) then
            RAM_rd_main <= '1';
            HASH_POS    <= HASH_POS+1;
            PTR_rd_main <= PTR_rd_main+1;
            intrn_count <= intrn_count + 1; 
           elsif (intrn_count = 8) then
            RAM_rd_main <= '1';
            HASH_POS    <= HASH_POS+1;
            PTR_rd_main <= PTR_rd_main+1;
            intrn_count <= intrn_count + 1; 
            HASH_OUT    <= '1';

            
                                 
  --          elsif(intrn_count = 9) then 
  --          intrn_count <= intrn_count + 1;
            elsif(intrn_count = 9) then 
            intrn_count <= (others =>'0');
            HASH_OUT    <= '0';
            RAM_rd_main <= '0';
            PTR_rd_main <= (others =>'0');      
        --    tag         <= "00";
          --  COUNT_OUT   <= COUNT_OUT +1; 
          --  tag         <= tag-1;
            if tag-1 <32 then 
               nxt_fsm <= LDMAIN;

            else 
               nxt_fsm <= INIT;

            end if; 
            end if;  
                

       when OUTPUT =>   nxt_fsm <= INIT;
    end case;
    end if; 
  end process cmb_fsm;
  
  
  OUT_PROC:process(clk,rst)
           begin 
           if (rst = '1') then 
                HASH_MSG <= (others =>'0'); 
           elsif(rising_edge(clk)) then 
               case conv_integer(HASH_POS) is 
               when 0 => HASH_MSG(7) <= ST_OP_main;
               when 1 => HASH_MSG(6) <= ST_OP_main;
               when 2 => HASH_MSG(5) <= ST_OP_main;
               when 3 => HASH_MSG(4) <= ST_OP_main;
               when 4 => HASH_MSG(3) <= ST_OP_main;
               when 5 => HASH_MSG(2) <= ST_OP_main;
               when 6 => HASH_MSG(1) <= ST_OP_main;
               when 7 => HASH_MSG(0) <= ST_OP_main;
               when others => HASH_MSG <= (others =>'Z');
               
               end case; 
           end if; 
           
           end process; 
  
 -- output state machine
 output_fsm: process(rst, clk)
  begin
    if (rst = '1') then
      reg_fsm_out  <= INIT_out;
    elsif (rising_edge(clk)) then
      reg_fsm_out <= nxt_fsm_out;
    end if;
  end process output_fsm; 
  
  cmb_fsm_output:process(clk,rst)
                 begin 
                    if(rst = '1') then 
                        OS_POS      <= (others =>'0');
                        PTR_wr_init <= (others =>'0');

                    elsif(rising_edge(clk))then 
                        case reg_fsm_out is
                        when INIT_OUT => 
                        if nxt_fsm = FINST_OUTPUT then 
                           nxt_fsm_out <= OUTOUT; 
                           PTR_wr_init <= (others =>'0');
                        elsif nxt_fsm = CLR_MAIN then 
                            if PTR_wr_init <231 then 
                               PTR_wr_init <=PTR_wr_init + 1; 
                            end if;
                        elsif nxt_fsm = INJECTMSG then 
                          if intrn_count = 0  then
                            PTR_wr_init <= (others =>'0');
                     --     elsif intrn_count = 4 then
                     --       PTR_wr_init <= (others =>'0');
                          elsif intrn_count = 5 then 
                            PTR_wr_init  <= PTR_wr_init+1;
                          elsif intrn_count = 7 then 
                            PTR_wr_init  <= (others =>'0');
                          end if; 
                        elsif nxt_fsm = LDMAIN then
                              PTR_wr_init <= (others =>'0');
                              OS_POS <= (X"00"); 
                              nxt_fsm_out <= SEC_OUT; 
                              PTR_wr_init <= (others =>'0');
                        end if;  
                              if PTR_wr_init >232 then 
                                PTR_wr_init <= (others=>'0');
                               end if; 
                        when SEC_OUT  =>
                            if QEN_OUT = '1' then 
                               OS_POS <= OS_POS + 1;
                            end if; 
                            if nxt_fsm = LDSEC then 
                                OS_POS <= (x"00");
                                
                                nxt_fsm_out <= MAIN_OUT; 
                              --  PTR_wr_init <= (others=>'0');
                            elsif nxt_fsm = LDMAIN then  
                                nxt_fsm_out <= SEC_OUT;
                              --  PTR_wr_init <= (others=>'0');
                            elsif nxt_fsm = FINST_OUTPUT then 
                               -- PTR_wr_init <= (others =>'0'); 
                                nxt_fsm_out <= INIT_OUT; 
                            end if; 
                               if PTR_wr_init >232 then 
                                PTR_wr_init <= (others=>'0');
                               end if; 
                            -- Flipping from state to another Ping-Pong Technology 
                        when MAIN_OUT =>
                            if QEN_OUT = '1' then 
                              if (LDR_INP = 31) then 
                               PTR_wr_init <= PTR_wr_init+1;
                              else 
                               OS_POS <= OS_POS + 1;
                              end if; 
                           
                            end if; 
                               if PTR_wr_init >232 then 
                                PTR_wr_init <= (others=>'0');
                               end if; 
                            if nxt_fsm = LDMAIN then 
                                OS_POS <= (x"00");
                                PTR_wr_init <=(others=>'0');
                                nxt_fsm_out <= SEC_OUT;
                            elsif nxt_fsm = LDSEC then  
                                nxt_fsm_out <= MAIN_OUT; 

                            elsif nxt_fsm = INJECTMSG then 
                                nxt_fsm_out <= INIT_OUT; 
                                PTR_wr_init <= (others=>'0');
                            elsif  nxt_fsm = FINST_OUTPUT then 
                                nxt_fsm_out <= OUTOUT;
                                PTR_wr_init <= (others=>'0');

                            end if; 
                        
                        when OUTOUT   =>
                             OS_POS      <= (others =>'0');
                             PTR_wr_init <= (others =>'0');
                             
                             if nxt_fsm = LDMAIN then 
                                nxt_fsm_out <= SEC_OUT;
                             elsif nxt_fsm = INIT then 
                                nxt_fsm_out <= INIT_OUT;
                             end if; 

                        when DONE     =>
                        end case; 
                    end if;
                 end process; 
  
  
  
 end Behavioral;
